#ifndef _material_h_
#define _material_h_
#include <math.h>

#ifndef _material_define_
#define _material_define_
//define material_type
const int Si     = 101;
const int GaAs   = 102;
const int Ge     = 103;
const int SiO2   = 111; 
const int Al     = 121;

const int semiconductor = 151;
const int metal         = 152;
const int Insulator     = 153; 

#endif


class Material
{
	private:
		double Eg300,EGALPH,EGBETA;     	//for ni
		double NC300,NC_F;              	//for Nc
		double NV300,NV_F;              	//for Nv
		double TAUN0,NSRHN,EXN_TAU;     	//for elec life
		double TAUP0,NSRHP,EXP_TAU;     	//for hole life
		double ETRAP,AUGN,AUGP,C_DIRECT;	//for recombination
		double MUN_MIN,MUN_MAX,NUN,XIN,NREFN,ALPHAN;
		double MUP_MIN,MUP_MAX,NUP,XIP,NREFP,ALPHAP;
		double EleMessConstant,HoleMessConstant;
		double ELE_TAUW,HOL_TAUW;
		
		double s_length,s_time,s_temp;
	public:	
		double eps0,eps;
		double eps_SiO2;
		double k,me,e;
	public:	
		Material(int ,double);	
	public:
		double  ni(double T)
		{
			double Eg = Eg300 + EGALPH*(1.0*1.0/(1.0 + EGBETA) - T*T/(T + EGBETA));
			return sqrt(Nc(T)*Nv(T))*exp(-Eg/(2*k*T));		
		}	 
		double  Vt(double T)
		{
			return k*T/e;		
		}	 
		double Vbi(double Na,double Nd,double T)
		{
			return k*T/e*asinh((Nd-Na)/(2*ni(T)));
		}	
		
		double Nc(double T)
		{
			return NC300*pow(T,NC_F);
		}     

		double Nv(double T)
		{
			return NV300*pow(T,NV_F);
		}  
 
		double ElecLiftTime(double Na,double Nd,double T)
		{
			return TAUN0/(1+(Na+Nd)/NSRHN)*pow(T,EXN_TAU);
		}

		double HoleLiftTime(double Na,double Nd,double T)
		{
			return TAUP0/(1+(Na+Nd)/NSRHP)*pow(T,EXP_TAU);
		}

		double recombination(double p,double n,double Na,double Nd,double T)
		{
			double nie = ni(T);
			double Ushr,Uauger,Udir;
			Ushr   = (p*n-nie*nie)/
			  	(HoleLiftTime(Na,Nd,T)*(n+nie*exp(ETRAP/(k*T)))+
			  	 ElecLiftTime(Na,Nd,T)*(p+nie*exp(-ETRAP/(k*T))));
			Uauger = AUGN*(p*n*n-n*nie*nie)+ AUGP*(p*p*n-p*nie*nie);
			Udir   = C_DIRECT*(p*n-nie*nie);
			return Ushr+Uauger+Udir;
		} 

		double mu0_elec(double Na,double Nd,double T)
		{
			return MUN_MIN + 
			       (MUN_MAX*pow(T,NUN)-MUN_MIN)/
			       (1+pow(T,XIN)*pow((Na+Nd)/NREFN,ALPHAN));
		}      

		double mu0_hole(double Na,double Nd,double T)
		{
			return MUP_MIN + 
			       (MUP_MAX*pow(T,NUP)-MUP_MIN)/
			       (1+pow(T,XIP)*pow((Na+Nd)/NREFP,ALPHAP));
		}
		
		double mu_elec(double Na,double Nd,double T)      //high field mobility	 
		{
			return mu0_elec(Na,Nd,T);
		}	

		double mu_hole(double Na,double Nd,double T)
		{
			return mu0_hole(Na,Nd,T);
		}			
		
		double mass_elec()
		{
			return EleMessConstant;
		}

		double mass_hole()
		{
			return HoleMessConstant;
		}

		double taop_elec(double Na,double Nd,double T)
		{
			return mass_elec()*mu_elec(Na,Nd,T)/e;
		}

		double taop_hole(double Na,double Nd,double T)
		{
			return mass_hole()*mu_hole(Na,Nd,T)/e;
		}

		double taow_elec()
		{
			return ELE_TAUW;
		}

		double taow_hole()
		{
			return HOL_TAUW;
		}

};	


#endif
