#ifndef _solver_h_
#define _solver_h_
#include "vec.h"
#include "petscksp.h"
#include "mesh.h"
#include "bc.h"
#include "material.h"

//define carrier type
const int N_Type  = 201;
const int P_Type  = 202;
const int PN_Both = 203;

//define solver type
const int DDM = 301;
const int HDM = 302;

//define flux function
const int Roe     = 401;
const int VanLeer = 402;
const int AUSM    = 403;

//define reconstruct order
const int FirstOrder  = 451;
const int SecondOrder = 452;
//define scheme
const int DDM_Gummel   = 501;
const int DDM_Newton   = 502;
const int HDM_explicit = 511;
const int HDM_implicit = 512;
//define solve type
const int EQUILIBRIUM =  601;
const int STEADYSTATE =  602;
const int DCSWEEP     =  603;
const int TRANSIENT   =  604;
//define parameter index
const int MeshObject  = 700;
const int ElecDensity = 701;
const int HoleDensity = 702;
const int EFieldX     = 703;
const int EFieldY     = 704;
const int Potential   = 705;
const int Doping      = 706;
const int Na          = 707;
const int Nd          = 708;
const int Temperature = 709;


const int Linear    = 751;
const int SignedLog = 752;

//define Newton's method
const int LineSearchCubic     = 855;
const int LineSearchQuadratic = 856;
const int LineSearchNo        = 857;
const int TrustRegion         = 860;

typedef struct
{
	int     Solver;
	int     Scheme;
	int     SNES;
	int     Type;
	int     FluxFunc;
  	int     Reconstruct;
  	double  CFL;
   	int     Electrode_VScan;
   	int     Electrode_Record;
  	char    IVFile[32];
    	double  VStart,VStep,VStop;
   	double  TStart,TStep,TStop;
  	double  AutoSave;
}SolveDefine;

typedef struct
{
	int     Variable;
	int     Measure;
  	double  Dispersion;
}RefineDefine;

typedef struct
{
	bool    core;
	bool    ascii;
  	char    CoreFile[32];
  	char    AscFile[32];
}ExtractDefine;

typedef struct
{
	bool    flag;
  	char    CoreFile[32];

}ImportDefine;

typedef struct
{
	char    Electrode[32];
	int     bc_index;
	vector<string> vsrc_name;
	vector<int>    vsrc_index;
}AttachDefine;

typedef struct
{
	int  MeshOnly;                   // draw mesh only (2D)
	int  Variable;                   // which Variable to plot
	int  Resolution;                 // x window resolution
  	int  Measure;                    // log or linear plot
  	int  GeneratePSFile;             // generate a postscript file
  	char PSFileName[32];
  	int  NoMesh;                     // do not plot the mesh
  	int  ConSurf;
  	int  NoSurf;
  	int  Style;                      // STYLE_BW | STYLE_WIREFRAME | STYLE_COLOR | (default) | STYLE_GRAYLEVEL
  	int  ztick;                      // # of z axis ticks
  	int  Persp;                      // Perspective distance

  	int  Con;                        // contour plot
  	int  NumCon;                     // # of contours

  	double az;                       // azimuthal angle
  	double el;                       // elevation angle
}PlotDefine;

class  SimData
{
   public:
 	double       	Na,Nd;
	vector_double 	Qn,Qp;
	double        	n,p;
	double        	P;
	double        	T;
	SimData()
	{
		Na=Nd=0; n=p=0; P=0; T=0;
		Set_Vector_zero(Qn);
		Set_Vector_zero(Qp);
	}
};

typedef  struct
{
 	double          Ex,Ey;
 	vector_double   Qnx,Qny,Qpx,Qpy;       //reconstruction grads
 	vector_double 	Qn_old,Qp_old;
	vector_double   *Qp_r,*Qn_r;           //reconstruction data,each node have nb_num elems
	vector_double   *Qp_s,*Qn_s;           //surface flux data,each node have nb_num elems
	vector_double   Qp_flux,Qn_flux;       //hydrodynamic flux
	vector_double   Rn_old,Rp_old;         //right hand size vector
}SimAuxData;

class BZoneData
{
  public:
	int             zone_index;
	ZONE            *pzone;
	int           	material_type;
  	int          	material;
  	int             node_num;
  	Material     	*mt;
  	SimData		*fs;
  	SimAuxData      *aux;
  	virtual   int   Init(ZONE*,char *,double,double)=0;
  	BZoneData()	{ mt = 0;fs = 0; aux = 0; }
  	virtual  ~BZoneData()	{}
};

//a semiconductor zone
class SMCZone: public BZoneData
{
  public:
	int		import_doping(char *,double);
	int	        export_doping(char *,double);
	int		import_solution(char *,char *,double);
	int	        export_solution(char *,char *,double);
  public:
  	//build the interface current between cell i and cell j
  	vector<int>     electrode;
  	double 		Jn_drift  (int ,int, double );
  	double 		Jn_diffuse(int ,int, double );
  	double 		Jp_drift  (int ,int, double );
  	double 		Jp_diffuse(int ,int, double );
    	int             get_electrode_current(DABC & ,int,double,double &);
  public:
        int             set_flux_function(int);

	void    	first_order_reconst(DABC &,int);    //first order reconstruction
	void    	MUSCL_reconst(DABC &,int);  	     //MUSCL type reconstruction, 2nd order
	//the pointer to flux_functions. set by set_flux_function
	vector_double   (*flux_function)(const vector_double &,const vector_double &,
				 double,double,double,double,double,double &);

	//build the numerical hydrodynamic flux of cell
	double          build_flux(DABC &,int);
	//get jacobian matrix of flux
	void            Get_jacobian_matrix(const vector_double & ,matrix_double &,matrix_double &,int);
	void            form_matrix(const vector_double &,const vector_double &,double ,double,
			            matrix_double& ,matrix_double&,int);
	void            build_right_const(int);
	void            build_implicit_equation(SolveDefine &,Mat&,Vec&,DABC &,double,int);
	void 		solution_update(Vec &,int);
	void            set_ohmic_boundary(DABC &,int);
	void            Flux_explicit_update(DABC &,int,double);
	//solution data synchronization
	void            sol_ddm_to_hdm();
	void            sol_hdm_to_ddm();
	//set source term
	void 		get_source_matrix(matrix_double &,int,int);
	void 		source_implicit_update(double);
	void 		source_explicit_update(double);
  public:
   	int             Init(ZONE*,char *,double,double);
   	void            report(){printf("This is a semiconductor Zone\n");}
  public:
  	~SMCZone()
  	{
    	    delete mt;
  	    delete [] fs;
  	    for(int i=0;i<node_num;i++)
  	    {
  	    	delete [] aux[i].Qp_r;
  	  	delete [] aux[i].Qn_r;
  	  	delete [] aux[i].Qp_s;
  	  	delete [] aux[i].Qn_s;
  	    }
  	    delete [] aux;
  	 }
};

//an insulator zone
class ISZone: public BZoneData
{
  public:
  	int		import_solution(char *,char *,double);
	int	        export_solution(char *,char *,double);
  public:
	int             Init(ZONE*,char *,double,double);
	void            report(){printf("This is an ISZone\n");}
  public:
   	~ISZone() 	{delete mt; delete [] fs;}
};


//an electrode zone
class ElZone: public BZoneData
{
  public:
  	int		import_solution(char *,char *,double);
	int	        export_solution(char *,char *,double);
  public:
	int             Init(ZONE*,char *,double,double);
	void            report(){printf("This is an ElZone\n");}
  public:
   	~ElZone() 	{delete mt; delete [] fs;}
};



class BSolver : public MESH
{
  public:
  	char 	    MeshFile[32];
	char 	    ModelFile[32];
	double      LatticeTemp;
	double      DeviceDepth;
	int         Carrier;
	vector<VSource*> vsrc;                     // dynamic array for vsource pointer

  	double    clock;
  	double    dt;
  	BZoneData **zonedata;
  	DABC      bc;
	int     static_config(list<CmdParseBuf> &);
	void    config_info();
  public:
  	void    reorder(); //reorder mesh and the data attached to mesh. for speed up implicit solver
  	double  set_scale(int,int,int,RefineDefine &);
  	int     refine(RefineDefine &);
  	int     read_mesh(char*);
  	int     build_zonedata();
   	int     setup_init_data(char*,double);
   	void    plot_data(PlotDefine &);
   	int     setup_breakrun_data(char*);
   	int     setup_bc(list<CmdParseBuf>&);
   	void    clear_data() {for(int i=0;i<zone_num;i++) delete zonedata[i]; delete [] zonedata;}
  public:
	BSolver():clock(0) {}
};

#endif

