#include "poisson.h"
#include <stdio.h>
#include <math.h>


void POISSON::init_poisson()
{
  int *nnz,*p;

  printf("Poisson solver init...");
  N = 0;
  //poisson is a global equation. all the nodes in each zone should be involved
  for(int i=0;i<zone_num;i++) N+=zone[i].davcell.size();

  nnz = new int[N];
  p = nnz;
  //pre-alloc approximate memory
  for(int i=0;i<zone_num;i++)
    for(int j=0;j<zone[i].davcell.size();j++)
         	*p++ = zone[i].davcell[j].nb_num+1;

  a1  = new double[N];
  a2  = new double[N];
  a3  = new double[N];
  for(int i=0;i<N;i++)   	a1[i]=a2[i]=a3[i]=0;
  //init least-squares parameters
  for(int i=0;i<zone_num;i++)
    for(int j=0; j<zone[i].davcell.size();j++)
  {
        int ofs = zone[i].offset;
       	const VoronoiCell *pcell = zone[i].davcell.GetPointer(j);

       	for(int k=0;k<pcell->nb_num;k++)
       	{
          const VoronoiCell *ncell = zone[i].davcell.GetPointer(pcell->nb_array[k]);
          double dx = ncell->x - pcell->x;
          double dy = ncell->y - pcell->y;
          double w  = 1.0/sqrt(dx*dx+dy*dy);
          a1[j+ofs]+=w*w*dx*dx;
          a2[j+ofs]+=w*w*dx*dy;
          a3[j+ofs]+=w*w*dy*dy;
        }

  }

  //init petse matrix and vector
  VecCreate(PETSC_COMM_WORLD,&x);
  VecSetSizes(x,PETSC_DECIDE,N);
  VecSetType(x, VECSEQ);
  VecDuplicate(x,&b);

  MatCreateSeqAIJ(PETSC_COMM_WORLD,N,N,0,nnz,&A);

  KSPCreate(PETSC_COMM_WORLD,&ksp);
  KSPSetOperators(ksp,A,A,SAME_PRECONDITIONER);
  KSPSetType(ksp,KSPPREONLY);
  //KSPSetType(ksp,KSPGMRES);
  KSPGetPC(ksp,&pc);
  //ierr = PCSetType(pc,PCILU);
  PCSetType(pc,PCLU);
  //PCLUSetDamping(pc,PETSC_DECIDE);

  //init matrix A.
  for(int i=0;i<zone_num;i++)
  {
        //------------------------------------------------------------------------------------------------
        //semiconductor zone
        if(zonedata[i]->material_type==semiconductor)
          for(int j=0;j<zone[i].davcell.size();j++)
	  {
               	const VoronoiCell *pcell = zone[i].davcell.GetPointer(j);

         	//inner point or neumann bc type
         	if(!pcell->bc_index||bc[pcell->bc_index-1].BCType==NeumannBoundary)
                 {
                 	double coefficient=0;
                       	for(int k=0;k<pcell->nb_num;k++)
                  	{
                  		double a=pcell->elen[k]/pcell->ilen[k]/pcell->area;
                  		MatSetValue(A,j+zone[i].offset,pcell->nb_array[k]+zone[i].offset,a,INSERT_VALUES);
        			coefficient+=-a;
        		}
        		MatSetValue(A,j+zone[i].offset,j+zone[i].offset,coefficient,INSERT_VALUES);
     			continue;
                 }

               	//first bc type
               	if(bc[pcell->bc_index-1].BCType==OhmicContract||bc[pcell->bc_index-1].BCType==SchottkyContract)
                  {MatSetValue(A,j+zone[i].offset,j+zone[i].offset,1,INSERT_VALUES); continue;}

                //third bc type
                if(bc[pcell->bc_index-1].BCType==InsulatorContract)
                {
                	double coefficient=0,a=0;
                	double eps_1 = zonedata[i]->mt->eps;
        		double eps_2 = zonedata[i]->mt->eps_SiO2;
        		double Thick = bc[pcell->bc_index-1].Thick;
                       	for(int k=0;k<pcell->nb_num;k++)
                  	{
                  		double a=pcell->elen[k]/pcell->ilen[k]/pcell->area;
                  		if(k==0||k==pcell->nb_num-1)
                  		{
                  			double bv=0.5*pcell->ilen[k]*(-0.25*eps_2/eps_1/Thick)/pcell->area;
                  			a+=bv;
                  			coefficient+=3*bv;
                  		}
                  		MatSetValue(A,j+zone[i].offset,pcell->nb_array[k]+zone[i].offset,a,INSERT_VALUES);
        			coefficient+=-a;
        		}

        		MatSetValue(A,j+zone[i].offset,j+zone[i].offset,coefficient,INSERT_VALUES);

                	continue;
                }

                //the boundary function: eps1*dP1/dn - eps2*dP2/dn = sigma
                if(bc[pcell->bc_index-1].BCType==InsulatorInterface)
                {
              		//first get norm vector to interface
              		double nx,ny;
              		for(int k=0;k<pcell->mirror_num;k++)
              		{
              			nx = pcell->mirror[k].nx;
              			ny = pcell->mirror[k].ny;
              			if(fabs(nx)+fabs(ny)>0.5) break;
              		}

              		//use lease-square technology to get derivative
              		double coefficient=0;
        		for(int k=0;k<pcell->nb_num;k++)
        		{
        			const VoronoiCell *ncell = zone[i].davcell.GetPointer(pcell->nb_array[k]);
        			int ofs = zone[i].offset;
        			double dx = ncell->x - pcell->x;
        			double dy = ncell->y - pcell->y;
        			double w=1.0/sqrt(dx*dx+dy*dy);
        			double ay = 1/(a1[j+ofs]*a3[j+ofs]-a2[j+ofs]*a2[j+ofs])*
        			           (a1[j+ofs]*w*w*dy-a2[j+ofs]*w*w*dx);
        			double ax = 1/(a1[j+ofs]*a3[j+ofs]-a2[j+ofs]*a2[j+ofs])*
        			           (a3[j+ofs]*w*w*dx-a2[j+ofs]*w*w*dy);
        			double eps = zonedata[i]->mt->eps;
        			double a = eps*(ax*nx + ay*ny);
        			MatSetValue(A,j+ofs,pcell->nb_array[k]+ofs,a,INSERT_VALUES);
        			coefficient+=-a;
        	  	}
        	  	MatSetValue(A,j+zone[i].offset,j+zone[i].offset,coefficient,INSERT_VALUES);

        		int donorzone,donornode;
            		zinterface.Find_donor(i,j,donorzone,donornode);
            		const VoronoiCell *dcell = zone[donorzone].davcell.GetPointer(donornode);
            		coefficient=0;
        		for(int k=0;k<dcell->nb_num;k++)
        		{
        			const VoronoiCell *ncell = zone[donorzone].davcell.GetPointer(dcell->nb_array[k]);
        			int ofs = zone[donorzone].offset;
        			double dx = ncell->x - dcell->x;
        			double dy = ncell->y - dcell->y;
        			double w=1.0/sqrt(dx*dx+dy*dy);
        			double ay = 1/(a1[donornode+ofs]*a3[donornode+ofs]-a2[donornode+ofs]*a2[donornode+ofs])*
        			           (a1[donornode+ofs]*w*w*dy-a2[donornode+ofs]*w*w*dx);
        			double ax = 1/(a1[donornode+ofs]*a3[donornode+ofs]-a2[donornode+ofs]*a2[donornode+ofs])*
        			           (a3[donornode+ofs]*w*w*dx-a2[donornode+ofs]*w*w*dy);
        			double eps = zonedata[donorzone]->mt->eps;
        			double a = eps*(ax*nx + ay*ny);
        			MatSetValue(A,j+zone[i].offset,dcell->nb_array[k]+ofs,-a,INSERT_VALUES);
        			coefficient+=-a;
        	  	}
        	  	MatSetValue(A,j+zone[i].offset,donornode+zone[donorzone].offset,-coefficient,INSERT_VALUES);
        	  	continue;

         	}
         }
        //------------------------------------------------------------------------------------------------
        //electrode zone
        if(zonedata[i]->material_type==metal)
          for(int j=0;j<zone[i].davcell.size();j++)
          {
               	const VoronoiCell *pcell = zone[i].davcell.GetPointer(j);
               	if(!pcell->bc_index||bc[pcell->bc_index-1].BCType==NeumannBoundary)
               	{
               		double coefficient=0;
               		for(int k=0;k<pcell->nb_num;k++)
               		{
               			double a=pcell->elen[k]/pcell->ilen[k]/pcell->area;
              			MatSetValue(A,j+zone[i].offset,pcell->nb_array[k]+zone[i].offset,a,INSERT_VALUES);
        			coefficient+=-a;
        		}
        		MatSetValue(A,j+zone[i].offset,j+zone[i].offset,coefficient,INSERT_VALUES);
     			continue;
     		}

     		if(bc[pcell->bc_index-1].BCType==OhmicContract||bc[pcell->bc_index-1].BCType==SchottkyContract)
                  {MatSetValue(A,j+zone[i].offset,j+zone[i].offset,1,INSERT_VALUES); continue;}

          }


       //------------------------------------------------------------------------------------------------
       //  Insulator zone
       if(zonedata[i]->material_type==Insulator)
          for(int j=0;j<zone[i].davcell.size();j++)
	  {

         	const VoronoiCell *pcell = zone[i].davcell.GetPointer(j);
        	//inner point or neumann bc type
         	if(!pcell->bc_index||bc[pcell->bc_index-1].BCType==NeumannBoundary)
                 {
                 	double coefficient=0;
                       	for(int k=0;k<pcell->nb_num;k++)
                  	{
                  		double a=pcell->elen[k]/pcell->ilen[k]/pcell->area;
                  		MatSetValue(A,j+zone[i].offset,pcell->nb_array[k]+zone[i].offset,a,INSERT_VALUES);
        			coefficient+=-a;
        		}
        		MatSetValue(A,j+zone[i].offset,j+zone[i].offset,coefficient,INSERT_VALUES);
     			continue;
                 }

        	if(bc[pcell->bc_index-1].BCType==InsulatorInterface)
              	{
              		int donorzone,donornode;
            		zinterface.Find_donor(i,j,donorzone,donornode);
            		MatSetValue(A,j+zone[i].offset,j+zone[i].offset,1,INSERT_VALUES);
            		MatSetValue(A,j+zone[i].offset,donornode+zone[donorzone].offset,-1,INSERT_VALUES);
            		continue;
            	}

            	if(bc[pcell->bc_index-1].BCType==GateContract)
              	{MatSetValue(A,j+zone[i].offset,j+zone[i].offset,1,INSERT_VALUES); continue;}

          }
   }


  MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
  MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);
  KSPSetUp(ksp);
  delete [] nnz ;
  printf("done\n");

}

void POISSON::do_poisson()
{
	double value;
	PetscScalar *data;

        for(int i=0;i<zone_num;i++)
        {
        //------------------------------------------------------------------------------------------------
        //semiconductor zone
        if(zonedata[i]->material_type==semiconductor)
          for(int j=0;j<zone[i].davcell.size();j++)
          {
 	        const VoronoiCell* pcell= zone[i].davcell.GetPointer(j);
 	        //inner point or NeumannBoundary
 	        if(!pcell->bc_index||bc[pcell->bc_index-1].BCType==NeumannBoundary)
 	        {
 	        	value = -1.0/zonedata[i]->mt->eps*(zonedata[i]->fs[j].Qp.v[0]-zonedata[i]->fs[j].Qn.v[0]+
 	               		zonedata[i]->fs[j].Nd-zonedata[i]->fs[j].Na);
 	               	VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	               	continue;
 	        }

 	        //first bc type
 	        if(pcell->bc_index&&bc[pcell->bc_index-1].BCType==OhmicContract)
 	        {
 	        	 double vapp = bc[pcell->bc_index-1].V;
 	        	 double K = zonedata[i]->mt->k;
 	        	 double T = zonedata[i]->fs[j].T;
 	        	 //the fermi potential of OhmicContract reference:dessis manual
 	        	 value = vapp+K*T*asinh((zonedata[i]->fs[j].Nd-zonedata[i]->fs[j].Na)/(2*zonedata[i]->mt->ni(T)));
 	        	 VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	        	 continue;
 	        }

	        if(pcell->bc_index&&bc[pcell->bc_index-1].BCType==SchottkyContract)
 	        {
 	        	 double vapp = bc[pcell->bc_index-1].V;
 	       	 	 double K = zonedata[i]->mt->k;
 	        	 double T = zonedata[i]->fs[j].T;
 	        	 //the potential in SchottkyContract
 	        	 value = vapp - bc[pcell->bc_index-1].Vbarrier
 	        	 	+ K*T*asinh((zonedata[i]->fs[j].Nd-zonedata[i]->fs[j].Na)/(2*zonedata[i]->mt->ni(T)));
	                 VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
	                 continue;
	        }
		 //third bc type
 	        if(pcell->bc_index&&bc[pcell->bc_index-1].BCType==InsulatorContract)
 	        {
 	        	 double vapp = bc[pcell->bc_index-1].V;

 	        	 double q = zonedata[i]->fs[j].Qp.v[0]-zonedata[i]->fs[j].Qn.v[0]+
 	               		    zonedata[i]->fs[j].Nd-zonedata[i]->fs[j].Na;
 	               	 double eps_1 = zonedata[i]->mt->eps;
        		 double eps_2 = zonedata[i]->mt->eps_SiO2;
 	          	 double r=q/eps_1 + eps_2/(eps_1*bc[pcell->bc_index-1].Thick)*vapp;

 	          	 value= -1.0/eps_1*q
 	          	 	-0.5*pcell->ilen[0]*r/pcell->area
 	          	 	-0.5*pcell->ilen[pcell->nb_num-1]*r/pcell->area;

 	          	 VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	               	 continue;

 	        }
 	        //the boundary function: eps1*dP1/dn - eps2*dP2/dn = sigma
 	        if(pcell->bc_index&&bc[pcell->bc_index-1].BCType==InsulatorInterface)
 	        {

 	       		value = zonedata[i]->fs[j].Qp.v[0]-zonedata[i]->fs[j].Qn.v[0]+
 	       			zonedata[i]->fs[j].Nd-zonedata[i]->fs[j].Na;
 	       		VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	       		continue;
 	       	}
	  }
	//------------------------------------------------------------------------------------------------
        //electrode zone
        if(zonedata[i]->material_type==metal)
          for(int j=0;j<zone[i].davcell.size();j++)
          {
               	const VoronoiCell *pcell = zone[i].davcell.GetPointer(j);
               	if(!pcell->bc_index||bc[pcell->bc_index-1].BCType==NeumannBoundary)
               	{
 	         	value = 0;
 	         	VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	         	continue;
 	        }

     		//first bc type
 	        if(pcell->bc_index&&bc[pcell->bc_index-1].BCType==OhmicContract)
 	        {
 	        	double vapp = bc[pcell->bc_index-1].V;
 	        	value = vapp;
 	        	VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	        	continue;
 	        }

	        if(pcell->bc_index&&bc[pcell->bc_index-1].BCType==SchottkyContract)
 	        {
 	        	 double vapp = bc[pcell->bc_index-1].V;
 	        	 value = vapp;
 	        	 VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	        	 continue;
	        }

          }


       //------------------------------------------------------------------------------------------------
       //  Insulator zone
       if(zonedata[i]->material_type==Insulator)
          for(int j=0;j<zone[i].davcell.size();j++)
	  {

         	const VoronoiCell *pcell = zone[i].davcell.GetPointer(j);
        	//inner point or neumann bc type
         	if(!pcell->bc_index||bc[pcell->bc_index-1].BCType==NeumannBoundary)
                {
                 	value = 0;
 	         	VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	         	continue;
                }

                if(pcell->bc_index&&bc[pcell->bc_index-1].BCType==InsulatorContract)
 	        {
 	        	 double vapp = bc[pcell->bc_index-1].V;

 	        	 double q = zonedata[i]->fs[j].Qp.v[0]-zonedata[i]->fs[j].Qn.v[0]+
 	               		    zonedata[i]->fs[j].Nd-zonedata[i]->fs[j].Na;
 	               	 double eps_1 = zonedata[i]->mt->eps;
        		 double eps_2 = zonedata[i]->mt->eps_SiO2;
 	          	 double r=q/eps_1 + eps_2/(eps_1*bc[pcell->bc_index-1].Thick)*vapp;

 	          	 value= -1.0/eps_1*q
 	          	 	-0.5*pcell->ilen[0]*r/pcell->area
 	          	 	-0.5*pcell->ilen[pcell->nb_num-1]*r/pcell->area;

 	          	 VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	               	 continue;

 	        }

        	if(bc[pcell->bc_index-1].BCType==InsulatorInterface)
              	{
              		value = 0;
 	               	VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	               	continue;
            	}

            	if(pcell->bc_index&&bc[pcell->bc_index-1].BCType==GateContract)
 	        {
 	          	double vapp = bc[pcell->bc_index-1].V;
 	          	value = vapp+bc[pcell->bc_index-1].WorkFunction;
 	               	VecSetValue(b,j+zone[i].offset,value,INSERT_VALUES);
 	               	continue;
 	        }


          }

        }


 	VecAssemblyBegin(b);
        VecAssemblyEnd(b);

        KSPSolve(ksp,b,x);

 	VecGetArray(x,&data);
	/*
	//calculate Ex ,Ey by integrating around the boundary of a closed control-volume:Green's theorem
	for(int i=0;i<zone_num;i++)
           for(int j=0; j<zone[i].davcell.size();j++)
	{
		const VoronoiCell *pcell = zone[i].davcell.GetPointer(j);
		double ux=0,uy=0;

		for(int k=0;k<pcell->nb_num;k++)
		{	ux += (data[pcell->nb_array[k]+zone[i].offset]+data[j+zone[i].offset])/2*
			      pcell->elen[k]*sin(pcell->angle[k]+3.1415926/2);
			uy -= (data[pcell->nb_array[k]+zone[i].offset]+data[j+zone[i].offset])/2*
			      pcell->elen[k]*cos(pcell->angle[k]+3.1415926/2);
		}
		if(pcell->bc_index)
		{
			const VoronoiCell *pcell_nb1 = zone[i].davcell.GetPointer(pcell->nb_array[0]);
			const VoronoiCell *pcell_nb2 = zone[i].davcell.GetPointer(pcell->nb_array[pcell->nb_num-1]);

			ux += data[j+zone[i].offset]*(pcell_nb1->y - pcell->y)/2;
			ux += data[j+zone[i].offset]*(pcell->y - pcell_nb2->y)/2;

			uy -= data[j+zone[i].offset]*(pcell_nb1->x - pcell->x)/2;
			uy -= data[j+zone[i].offset]*(pcell->x - pcell_nb2->x)/2;
		}
		zonedata[i]->aux[j].Ex = -ux/pcell->area;
		zonedata[i]->aux[j].Ey = -uy/pcell->area;
		zonedata[i]->fs[j].P = data[j+zone[i].offset];
	}
	*/
	//calculate Ex and Ey by least-squares gradient construction
        for(int i=0;i<zone_num;i++)
           for(int j=0; j<zone[i].davcell.size();j++)
            {
        	double a=0,b=0,c=0,d=0,e=0,w=0;
        	const VoronoiCell *pcell = zone[i].davcell.GetPointer(j);

        	for(int k=0;k<pcell->nb_num;k++)
        	{
        		const VoronoiCell *ncell = zone[i].davcell.GetPointer(pcell->nb_array[k]);
        		double dx = ncell->x - pcell->x;
        		double dy = ncell->y - pcell->y;
        		double du = data[pcell->nb_array[k]+zone[i].offset]-data[j+zone[i].offset];
        		w=1.0/sqrt(dx*dx+dy*dy);
        		a+=w*w*dx*dx;
        		b+=w*w*dx*dy;
        	 	c+=w*w*dy*dy;
        	 	d+=w*w*du*dx;
        	 	e+=w*w*du*dy;
        	 }
        	 double ux=(c*d-b*e)/(a*c-b*b);
        	 double uy=(a*e-b*d)/(a*c-b*b);
        	 zonedata[i]->aux[j].Ex = -ux;
		 zonedata[i]->aux[j].Ey = -uy;
		 zonedata[i]->fs[j].P = data[j+zone[i].offset];

	  }

	VecRestoreArray(x,&data);

}


void POISSON::destroy_poisson()
{
	N = 0;
	delete [] a1;
  	delete [] a2;
 	delete [] a3;
 	VecDestroy(x);
 	VecDestroy(b);
   	MatDestroy(A);
   	KSPDestroy(ksp);
}
